"""An easy-to-use ML framework."""
__version__ = '0.5b1'

import os
import logging

# 配置默认环境变量
os.environ.setdefault('CLASSICML_ENGINE', 'Python')
os.environ.setdefault('CLASSICML_FONT', 'Arial Unicode MS')
# 系统logger
logging.basicConfig(level=logging.INFO)
CLASSICML_LOGGER = logging.getLogger(name='classicML')
CLASSICML_LOGGER.info('正在使用 {} 引擎'.format(os.environ['CLASSICML_ENGINE']))

from classicML.api import models
from classicML.api import plots

from classicML.api import BackPropagationNeuralNetwork
from classicML.api import BPNN
from classicML.api import DecisionTreeClassifier
from classicML.api import LogisticRegression
from classicML.api import LinearDiscriminantAnalysis
from classicML.api import LDA
from classicML.api import RadialBasisFunctionNetwork
from classicML.api import RBF
from classicML.api import SupportVectorClassifier
from classicML.api import SVC

from classicML.backend import activations
from classicML.backend import callbacks
from classicML.backend import initializers
from classicML.backend import kernels
from classicML.backend import losses
from classicML.backend import metrics
from classicML.backend import optimizers

from classicML import benchmarks


import numpy as np

from classicML.api.plots import _plt as plt
from classicML.api.plots.utils import _set_svc_axis
from classicML.api.plots.utils import _svc_plot_config
from classicML import CLASSICML_LOGGER


def _plot_sample_and_support_scatter(x, y, support):
    """绘制样本点和支持向量.

    Arguments:
        x: numpy.ndarray, array-like, 特征数据.
        y: numpy.ndarray, array-like, 标签.
        support: numpy.ndarray, 分类器的支持向量下标数组.
    """
    negative_label = [y == -1][0]
    positive_label = [y == 1][0]

    plt.scatter(x[negative_label, 0], x[negative_label, 1], c='lightcoral', marker='o', label='反例')
    plt.scatter(x[positive_label, 0], x[positive_label, 1], c='c', marker='o', label='正例')
    plt.scatter(x[support, 0], x[support, 1], c='white', marker='o', edgecolors='green', s=150, alpha=0.3)


def _plot_decision_boundary(svc):
    """绘制决策边界.

    Arguments:
        svc: classicML.models.SVC, 支持向量分类器实例.

    Returns:
        matplotlib的轮廓集.
    """
    # 生成向量空间.
    x0_coord = np.linspace(0, 1, 300)
    x1_coord = np.linspace(0, 0.8, 300)
    vector_matrix_0, vector_matrix_1 = np.meshgrid(x0_coord, x1_coord)

    # 进行预测并改变形状.
    y_pred = svc.predict(np.c_[vector_matrix_0.ravel(), vector_matrix_1.ravel()])
    y_pred = y_pred.reshape(vector_matrix_0.shape)

    # 使用向量空间的点进行绘制决策边界(y_pred是到平面的高度, 代替直接绘制y=0)
    CS = plt.contour(vector_matrix_0, vector_matrix_1, y_pred, [0], colors='orange', linewidths=1)

    return CS


def plot_support_vector_classifier(svc, x, y, x_label=None, y_label=None):
    """支持向量分类器二维示意图.

    Arguments:
        svc: classicML.models.SVC, 支持向量分类器实例.
        x: numpy.ndarray, array-like, 特征数据.
        y: numpy.ndarray, array-like, 标签.
        x_label: str, default=None,
            横轴的标签.
        y_label: str, default=None,
            纵轴的标签.
    """
    if svc.support is None:
        CLASSICML_LOGGER.error('模型没有训练')
        raise ValueError('你必须先进行训练')

    ax = plt.subplot()
    _set_svc_axis(ax)

    # 绘制样本点和支持向量
    _plot_sample_and_support_scatter(x, y, svc.support)

    # 绘制决策边界
    CS = _plot_decision_boundary(svc)

    _svc_plot_config(CS, svc.kernel.name, svc.C, x_label, y_label)
    plt.show()


import os

from classicML import CLASSICML_LOGGER

from classicML.backend.python import activations
from classicML.backend.python import callbacks
from classicML.backend.python import initializers
from classicML.backend.python import kernels
from classicML.backend.python import losses
from classicML.backend.python import metrics
from classicML.backend.python import optimizers
from classicML.backend.python import tree

if os.environ['CLASSICML_ENGINE'] == 'CC':
    from classicML.backend.cc.ops import __version__
    from classicML.backend.cc.ops import cc_calculate_error
    from classicML.backend.cc.ops import cc_calculate_error as calculate_error
    from classicML.backend.cc.ops import cc_clip_alpha
    from classicML.backend.cc.ops import cc_clip_alpha as clip_alpha
    from classicML.backend.cc.ops import cc_get_w
    from classicML.backend.cc.ops import cc_get_w as get_w
    from classicML.backend.cc.ops import cc_get_within_class_scatter_matrix
    from classicML.backend.cc.ops import cc_get_within_class_scatter_matrix as get_within_class_scatter_matrix
    from classicML.backend.cc.ops import cc_select_second_alpha
    from classicML.backend.cc.ops import cc_select_second_alpha as select_second_alpha
    from classicML.backend.cc.ops import cc_type_of_target
    from classicML.backend.cc.ops import cc_type_of_target as type_of_target
    CLASSICML_LOGGER.info('后端版本是: {}'.format(__version__))
else:
    from classicML.backend.python.ops import calculate_error
    from classicML.backend.python.ops import clip_alpha
    from classicML.backend.python.ops import get_w
    from classicML.backend.python.ops import get_within_class_scatter_matrix
    from classicML.backend.python.ops import select_second_alpha
    from classicML.backend.python.ops import type_of_target

from classicML.backend.training import get_initializer
from classicML.backend.training import get_kernel
from classicML.backend.training import get_loss
from classicML.backend.training import get_metric
from classicML.backend.training import get_optimizer
from classicML.backend.training import get_pruner


from classicML import CLASSICML_LOGGER
from classicML.backend import initializers
from classicML.backend import kernels
from classicML.backend import losses
from classicML.backend import optimizers
from classicML.backend import metrics
from classicML.backend.python import tree


def get_initializer(initializer, seed):
    """获取使用的初始化器实例.

    Arguments:
        initializer: str or classicML.initializers.Initializer 实例,
            初始化器.
        seed: int, 初始化器的随机种子.

    Raises:
        AttributeError: 模型编译的参数输入错误.
    """
    if isinstance(initializer, str):
        if initializer == 'random_normal':
            return initializers.RandomNormal(seed=seed)
        elif initializer == 'he_normal':
            return initializers.HeNormal(seed=seed)
        elif initializer in ('xavier_normal', 'glorot_normal'):
            return initializers.XavierNormal(seed=seed)
        elif initializer == 'rbf_normal':
            return initializers.RBFNormal(seed=seed)
        else:
            CLASSICML_LOGGER.error('初始化器调用错误')
            raise AttributeError
    elif isinstance(initializer, initializers.Initializer):
        return initializer
    elif initializer is None:
        return initializers.RandomNormal(seed=seed)
    else:
        CLASSICML_LOGGER.error('初始化器调用错误')
        raise AttributeError


def get_loss(loss):
    """获取使用的损失函数实例.

    Arguments:
        loss: str or classicML.losses.Loss 实例,
            损失函数.
    """
    if isinstance(loss, str):
        if loss in ('mse', 'mean_squared_error'):
            return losses.MeanSquaredError()
        elif loss == 'log_likelihood':
            return losses.LogLikelihood()
        elif loss == 'binary_crossentropy':
            return losses.BinaryCrossentropy()
        elif loss == 'categorical_crossentropy':
            return losses.CategoricalCrossentropy()
        elif loss == 'crossentropy':
            return losses.Crossentropy()
        else:
            CLASSICML_LOGGER.warn('你没有输入损失函数或者输入的损失函数不正确, 将使用默认的损失函数')
            return losses.Crossentropy()
    elif isinstance(loss, losses.Loss):
        return loss
    else:
        CLASSICML_LOGGER.warn('你没有输入损失函数或者输入的损失函数不正确, 将使用默认的损失函数')
        return losses.Crossentropy()


def get_optimizer(optimizer):
    """获取使用的优化器实例.

    Arguments:
        optimizer: str or classicML.optimizers.Optimizer 实例,
            优化器.

    Raises:
        AttributeError: 模型编译的参数输入错误.
    """
    if isinstance(optimizer, str):
        if optimizer in ('gd', 'gradient_descent'):
            return optimizers.GradientDescent()
        elif optimizer == 'newton':
            return optimizers.NewtonMethod()
        elif optimizer in ('sgd', 'stochastic_gradient_descent'):
            return optimizers.StochasticGradientDescent()
        elif optimizer == 'adam':
            return optimizers.Adam()
        elif optimizer == 'rbf':
            return optimizers.RadialBasisFunctionOptimizer()
        elif optimizer == 'SMO':
            return optimizers.SequentialMinimalOptimization()
        else:
            CLASSICML_LOGGER.error('优化器调用错误')
            raise AttributeError
    elif isinstance(optimizer, optimizers.Optimizer):
        return optimizer
    else:
        CLASSICML_LOGGER.error('优化器调用错误')
        raise AttributeError


def get_metric(metric):
    """获取使用的评估函数实例.

    Arguments:
        metric: str or classicML.metrics.Metric 实例,
            评估函数.

    Raises:
        AttributeError: 模型编译的参数输入错误.
    """
    if isinstance(metric, str):
        if metric == 'binary_accuracy':
            return metrics.BinaryAccuracy()
        elif metric == 'categorical_accuracy':
            return metrics.CategoricalAccuracy()
        elif metric == 'accuracy':
            return metrics.Accuracy()
    elif isinstance(metric, metrics.Metric):
        return metric
    else:
        CLASSICML_LOGGER.error('评估函数调用错误')
        raise AttributeError


def get_kernel(kernel, gamma):
    """获取使用的核函数实例.

    Arguments:
        kernel: str or classicML.kernels.Kernel 实例,
            核函数.
        gamma: float, 核函数系数.

    Raises:
        AttributeError: 模型编译的参数输入错误.
    """
    if isinstance(kernel, str):
        if kernel == 'linear':
            return kernels.Linear()
        elif kernel == 'rbf':
            return kernels.RBF(gamma=gamma)
        elif kernel == 'gaussian':
            return kernels.Gaussian(gamma=gamma)
        elif kernel == 'poly':
            return kernels.Polynomial(gamma=gamma)
        elif kernel == 'sigmoid':
            return kernels.Sigmoid(gamma=gamma)
        else:
            CLASSICML_LOGGER.error('核函数调用错误')
            raise AttributeError
    elif isinstance(kernel, kernels.Kernel):
        return kernel
    else:
        CLASSICML_LOGGER.error('核函数调用错误')
        raise AttributeError


def get_criterion(criterion):
    """获取使用的划分选择方式.

    Arguments:
        criterion: str,
            决策树学习的划分方式.

    Raises:
        AttributeError: 选择错误.
    """
    if criterion == 'gini':
        return tree.criterions.Gini()
    elif criterion == 'gain':
        return tree.criterions.Gain()
    elif criterion == 'entropy':
        return tree.criterions.Entropy()
    else:
        CLASSICML_LOGGER.error('选择错误')
        raise AttributeError


def get_pruner(pruning):
    """获取剪枝器.

    Arguments:
        pruning: str,
            决策树剪枝的方式.

    Raises:
        AttributeError: 选择错误.
    """
    if pruning == 'post':
        return tree.pruners.PostPruner()
    elif pruning == 'pre':
        return tree.pruners.PrePruner()
    elif pruning is None:
        return None
    else:
        CLASSICML_LOGGER.error('选择错误')
        raise AttributeError


import os
import time

import psutil

from classicML import CLASSICML_LOGGER


def timer(function):
    """程序计时装饰器.
    """
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()  # 注意将记录time.sleep()的时间
        return_values = function(*args, **kwargs)
        end_time = time.perf_counter()
        CLASSICML_LOGGER.info('耗时 {:.5f} s'.format(end_time - start_time))

        return return_values

    return wrapper


def memory_monitor(function):
    """内存监视装饰器.

    Notes:
        - 使用该函数统计内存信息, 有潜在降低运行速度的可能性.
        并且psutil针对的Python优化手段会导致在CC引擎的速度大幅降低.
    """
    def wrapper(*args, **kwargs):
        return_values = function(*args, **kwargs)

        pid = os.getpid()
        current_process = psutil.Process(pid)
        process_memory = current_process.memory_full_info()
        CLASSICML_LOGGER.info('占用内存 {:.5f} MB'.format(process_memory.uss / 1024 / 1024))
        return return_values

    return wrapper


import numpy as np
import pandas as pd


class Pruner(object):
    """剪枝器基类."""
    def __init__(self, name=None):
        super(Pruner, self).__init__()
        self.name = name

    def __call__(self, x, y, x_validation, y_validation, tree):
        """进行剪枝操作.

        Arguments:
            x: pandas.DataFrame, 特征数据.
            y: pandas.DataFrame, 标签.
            x_validation: pandas.DataFrame, 剪枝使用的验证特征数据.
            y_validation: pandas.DataFrame, 剪枝使用的验证标签.
            tree: classicML.backend.tree._TreeNode实例,
                决策树.
        """
        raise NotImplementedError

    def calculation_accuracy(self, *args, **kwargs):
        """计算使用预后剪枝操作的之后前的准确率."""
        raise NotImplementedError


class PostPruner(Pruner):
    """后剪枝器.
    """
    def __init__(self, name='post'):
        super(PostPruner, self).__init__(name=name)

    def __call__(self, x, y, x_validation, y_validation, tree):
        """进行剪枝操作.

        Arguments:
            x: pandas.DataFrame, 特征数据.
            y: pandas.DataFrame, 标签.
            x_validation: pandas.DataFrame, 剪枝使用的验证特征数据.
            y_validation: pandas.DataFrame, 剪枝使用的验证标签.
            tree: classicML.backend.tree._TreeNode实例,
                决策树.
        """
        if tree.leaf:
            return tree

        if x_validation.empty:
            return tree

        # TODO(Steve R. Sun): 函数value_counts在两类元素值相同时, 不会按照某种顺序排序,
        # 而是随机返回, 因此可能遇到决策树生成不一样的情况. 应该重写此函数.
        current_category = pd.value_counts(y).index[0]
        accuracy_after_pruning = np.mean(y_validation == current_category)

        if tree.continuous:
            # 使用二分法对连续值进行处理.
            greater_part_of_train = (x.loc[:, tree.feature_name] >= tree.dividing_point)
            less_part_of_train = (x.loc[:, tree.feature_name] < tree.dividing_point)
            greater_part_of_validation = (x_validation.loc[:, tree.feature_name] >= tree.dividing_point)
            less_part_of_validation = (x_validation.loc[:, tree.feature_name] < tree.dividing_point)

            greater_part_tree = self.__call__(x[greater_part_of_train],
                                              y[greater_part_of_train],
                                              x_validation[greater_part_of_validation],
                                              y_validation[greater_part_of_validation],
                                              tree.subtree['>= {:.3f}'.format(tree.dividing_point)])
            less_part_tree = self.__call__(x[less_part_of_train],
                                           y[less_part_of_train],
                                           x_validation[less_part_of_validation],
                                           y_validation[less_part_of_validation],
                                           tree.subtree['< {:.3f}'.format(tree.dividing_point)])
            tree.subtree['>= {:.3f}'.format(tree.dividing_point)] = greater_part_tree
            tree.subtree['< {:.3f}'.format(tree.dividing_point)] = less_part_tree

            # 重新计算叶子数目和高度.
            tree.num_of_leaves = (greater_part_tree.num_of_leaves + less_part_tree.num_of_leaves)
            tree.height = max(greater_part_tree.height, less_part_tree.height) + 1
            # 此结点是一个分支结点且在最底层, 需要进行处理.
            if greater_part_tree.leaf and less_part_tree.leaf:
                accuracy_before_pruning = self.calculation_accuracy(x_validation, y_validation, tree)
                if accuracy_after_pruning > accuracy_before_pruning:
                    tree.reset(pd.value_counts(y).index[0])
        else:
            max_height = -1
            tree.num_of_leaves = 0
            all_leaves = True

            for key in tree.subtree.keys():
                part_of_train = (x.loc[:, tree.feature_name] == key)
                part_of_validation = (x_validation.loc[:, tree.feature_name] == key)

                tree.subtree[key] = self.__call__(x[part_of_train],
                                                  y[part_of_train],
                                                  x_validation[part_of_validation],
                                                  y_validation[part_of_validation],
                                                  tree.subtree[key])

                # 重新计算叶子数目和高度.
                tree.num_of_leaves += tree.subtree[key].num_of_leaves
                if tree.subtree[key].height > max_height:
                    max_height = tree.subtree[key].height + 1

                if tree.subtree[key].leaf is False:
                    all_leaves = False

            tree.height = max_height

            # 此结点是一个分支结点且在最底层, 需要进行处理.
            if all_leaves:
                accuracy_before_pruning = self.calculation_accuracy(x_validation, y_validation, tree)
                if accuracy_after_pruning > accuracy_before_pruning:
                    tree.reset(pd.value_counts(y).index[0])

        return tree

    def calculation_accuracy(self, x_validation, y_validation, tree):
        """计算剪枝前的准确率.
        这里没有采取原文的做法, 而是只计算这一个分支的数据,
        因为不修剪其他分支, 其他分支当前的值不改变也就不会影响准确率的总体变化,
        这样不仅代码好实现, 而且同时显著减少计算的开销.

        Arguments:
            x_validation: pandas.DataFrame, 剪枝使用的验证特征数据.
            y_validation: pandas.DataFrame, 剪枝使用的验证标签.
            tree: classicML.backend.tree._TreeNode实例,
                决策树.
        """
        if tree.dividing_point is None:
            part_of_validation = x_validation.loc[:, tree.feature_name]
        else:
            def _divide(feature):
                """对连续值的样本进行二分组."""
                if feature >= tree.dividing_point:
                    return '>= {:.3f}'.format(tree.dividing_point)
                else:
                    return '< {:.3f}'.format(tree.dividing_point)

            part_of_validation = x_validation.loc[:, tree.feature_name].map(_divide)

        correct_samples = y_validation.groupby(part_of_validation).apply(
            lambda x: np.sum(x == tree.subtree[x.name].category)
        )

        accuracy = correct_samples.sum() / y_validation.shape[0]

        return accuracy


class PrePruner(Pruner):
    """预剪枝器.

    Notes:
        - 这里只取用了预剪枝算法的思想, 实际实现还是在决策树生成以后进行的剪枝操作,
          因为如果按照原文实现势必影响一个正常的决策树生成.
    """
    def __init__(self, name='pre'):
        super(PrePruner, self).__init__(name=name)

    def __call__(self, x, y, x_validation, y_validation, tree):
        """进行剪枝操作.

        Arguments:
            x: pandas.DataFrame, 特征数据.
            y: pandas.DataFrame, 标签.
            x_validation: pandas.DataFrame, 剪枝使用的验证特征数据.
            y_validation: pandas.DataFrame, 剪枝使用的验证标签.
            tree: classicML.backend.tree._TreeNode实例,
                决策树.
        """
        if tree.leaf:
            return tree

        if x_validation.empty:
            return tree

        # TODO(Steve R. Sun): 函数value_counts在两类元素值相同时, 不会按照某种顺序排序,
        # 而是随机返回, 因此可能遇到决策树生成不一样的情况. 应该重写此函数.
        current_category = pd.value_counts(y).index[0]
        accuracy_before_pruning = np.mean(y_validation == current_category)

        if tree.continuous:
            accuracy_after_pruning = self.calculation_accuracy(x[tree.feature_name],
                                                               y,
                                                               x_validation[tree.feature_name],
                                                               y_validation,
                                                               tree)
            # 泛化性能没有提高不进行剪枝, 故设置为叶结点.
            if accuracy_before_pruning >= accuracy_after_pruning:
                tree.reset(pd.value_counts(y).index[0])
            else:
                # 使用二分法对连续值进行处理.
                greater_part_of_train = (x.loc[:, tree.feature_name] >= tree.dividing_point)
                less_part_of_train = (x.loc[:, tree.feature_name] < tree.dividing_point)
                greater_part_of_validation = (x_validation.loc[:, tree.feature_name] >= tree.dividing_point)
                less_part_of_validation = (x_validation.loc[:, tree.feature_name] < tree.dividing_point)

                less_part_tree = self.__call__(x[less_part_of_train],
                                               y[less_part_of_train],
                                               x_validation[less_part_of_validation],
                                               y_validation[less_part_of_validation],
                                               tree.subtree['< {:.3f}'.format(tree.dividing_point)])
                greater_part_tree = self.__call__(x[greater_part_of_train],
                                                  y[greater_part_of_train],
                                                  x_validation[greater_part_of_validation],
                                                  y_validation[greater_part_of_validation],
                                                  tree.subtree['>= {:.3f}'.format(tree.dividing_point)])
                tree.subtree['< {:.3f}'.format(tree.dividing_point)] = less_part_tree
                tree.subtree['>= {:.3f}'.format(tree.dividing_point)] = greater_part_tree

                # 重新计算叶子数目和高度.
                tree.height = max(greater_part_tree.height, less_part_tree.height) + 1
                tree.num_of_leaves = (greater_part_tree.num_of_leaves + less_part_tree.num_of_leaves)
        else:
            accuracy_after_pruning = self.calculation_accuracy(x[tree.feature_name],
                                                               y,
                                                               x_validation[tree.feature_name],
                                                               y_validation,
                                                               tree)
            # 泛化性能没有提高不进行剪枝, 故设置为叶结点.
            if accuracy_before_pruning >= accuracy_after_pruning:
                tree.reset(pd.value_counts(y).index[0])
            else:
                max_height = -1
                tree.num_of_leaves = 0

                for key in tree.subtree.keys():
                    part_of_train = (x.loc[:, tree.feature_name] == key)
                    part_of_validation = (x_validation.loc[:, tree.feature_name] == key)

                    tree.subtree[key] = self.__call__(x[part_of_train],
                                                      y[part_of_train],
                                                      x_validation[part_of_validation],
                                                      y_validation[part_of_validation],
                                                      tree.subtree[key])

                    # 重新计算叶子高度和数目.
                    if tree.subtree[key].height > max_height:
                        max_height = tree.subtree[key].height + 1
                    tree.num_of_leaves += tree.subtree[key].num_of_leaves

                tree.height = max_height + 1

        return tree

    def calculation_accuracy(self, x, y, x_validation, y_validation, tree):
        """计算预剪枝划分后的准确率.

        Arguments:
            x: pandas.DataFrame, 特征数据.
            y: pandas.DataFrame, 标签.
            x_validation: pandas.DataFrame, 剪枝使用的验证特征数据.
            y_validation: pandas.DataFrame, 剪枝使用的验证标签.
            tree: classicML.backend.tree._TreeNode实例,
                决策树.
        """
        # 筛选出分支结点属性对应的属性值.
        if tree.dividing_point:
            def _divide(feature):
                """对连续值的样本进行二分组."""
                if feature >= tree.dividing_point:
                    return '>= {:.3f}'.format(tree.dividing_point)
                else:
                    return '< {:.3f}'.format(tree.dividing_point)

            part_of_train = x.loc[:, tree.feature_name].map(_divide)
            part_of_validation = x_validation.loc[:, tree.feature_name].map(_divide)
        else:
            part_of_train = x
            part_of_validation = x_validation

        # 获取分支结点每个属性标记最多的类别.
        category_in_train_set = y.groupby(part_of_train).apply(lambda x_: pd.value_counts(x_).index[0])

        correct_samples = y_validation.groupby(part_of_validation).apply(
            lambda x_: np.sum(x_ == category_in_train_set[x_.name])
        )

        accuracy = correct_samples.sum() / y_validation.shape[0]

        return accuracy


import numpy as np

from classicML import CLASSICML_LOGGER
from classicML.backend import get_within_class_scatter_matrix
from classicML.backend import get_w


class LinearDiscriminantAnalysis(object):
    """线性判别分析.

    Attributes:
        w: numpy.ndarray, 投影向量.
        mu_0: numpy.ndarray, 反例的均值.
        mu_1: numpy.ndarray, 正例的均值.
    """
    def __init__(self):
        super(LinearDiscriminantAnalysis, self).__init__()
        self.w = None
        self.mu_0 = None
        self.mu_1 = None

    def fit(self, x, y):
        """训练模型.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.

        Returns:
            LDA实例.
        """
        x = np.asarray(x)
        y = np.asarray(y)

        X_0 = x[y == 0]
        X_1 = x[y == 1]

        self.mu_0 = np.mean(X_0, axis=0, keepdims=True)
        self.mu_1 = np.mean(X_1, axis=0, keepdims=True)

        # 获得类内散度矩阵
        S_w = get_within_class_scatter_matrix(X_0, X_1, self.mu_0, self.mu_1)
        # 获得投影向量
        self.w = get_w(S_w, self.mu_0, self.mu_1)

        return self

    def predict(self, x):
        """模型进行预测.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.

        Returns:
            LDA预测的标签张量.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.w is None:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        coord = np.dot(x, self.w.T)

        center_0 = np.dot(self.w, self.mu_0.T)
        center_1 = np.dot(self.w, self.mu_1.T)

        y_pred = np.abs(coord - center_0) > np.abs(coord - center_1)
        y_pred = np.squeeze(y_pred)
        y_pred = y_pred.astype(int)

        return y_pred


import numpy as np


class Kernel(object):
    """核函数的基类.

    Attributes:
        name: str, default=None,
            核函数名称.

    Raises:
       NotImplementedError: __call__方法需要用户实现.
    """
    def __init__(self, name=None):
        """
        Arguments:
            name: str, default=None,
                核函数名称.
        """
        self.name = name

    def __call__(self, x_i, x_j):
        raise NotImplementedError


class Linear(Kernel):
    """线性核函数.
    """
    def __init__(self, name='linear'):
        super(Linear, self).__init__(name=name)

    def __call__(self, x_i, x_j):
        """
        Arguments:
            x_i: numpy.ndarray, 第一组特征向量.
            x_j: numpy.ndarray, 第二组特征向量.

        Returns:
            核函数映射后的特征向量.
        """
        kappa = np.matmul(x_j, x_i.T)

        return np.asmatrix(kappa)


class RBF(Kernel):
    """径向基核函数.

    Attributes:
        name: str, default='rbf',
            核函数名称.
        gamma: float, default=1.0,
            核函数系数.
    """
    def __init__(self, name='rbf', gamma=1.0):
        super(RBF, self).__init__(name=name)

        self.gamma = gamma

    def __call__(self, x_i, x_j):
        """
        Arguments:
            x_i: numpy.ndarray, 第一组特征向量.
            x_j: numpy.ndarray, 第二组特征向量.

        Returns:
            核函数映射后的特征向量.
        """
        kappa = np.exp(self.gamma * -np.sum(np.power(x_j - x_i, 2), axis=1))

        return np.asmatrix(kappa)


class Gaussian(RBF):
    """高斯核函数.
        具体实现参看径向基核函数.
    """
    def __init__(self, name='gaussian', gamma=1.0):
        super(Gaussian, self).__init__(name=name, gamma=gamma)


class Polynomial(Kernel):
    """多项式核函数.

    Attributes:
        name: str, default='rbf',
            核函数名称.
        gamma: float, default=1.0,
            核函数系数.
        degree: int, default=3,
            多项式的次数.
    """
    def __init__(self, name='poly', gamma=1.0, degree=3):
        super(Polynomial, self).__init__(name=name)

        self.gamma = gamma
        self.degree = degree

    def __call__(self, x_i, x_j):
        """
        Arguments:
            x_i: numpy.ndarray, 第一组特征向量.
            x_j: numpy.ndarray, 第二组特征向量.

        Returns:
            核函数映射后的特征向量.
        """
        kappa = self.gamma * np.power(np.matmul(x_j, x_i.T), self.degree)

        return np.asmatrix(kappa)


class Sigmoid(Kernel):
    """Sigmoid核函数.

    Attributes:
        name: str, default='rbf',
            核函数名称.
        gamma: float, default=1.0,
            核函数系数.
        beta: float, default=1.0,
            核函数参数.
        theta: float, default=-1.0,
            核函数参数.
    """
    def __init__(self, name='sigmoid', gamma=1.0, beta=1.0, theta=-1.0):
        super(Sigmoid, self).__init__(name=name)

        self.gamma = gamma
        self.beta = beta
        self.theta = theta

    def __call__(self, x_i, x_j):
        """
        Arguments:
            x_i: numpy.ndarray, 第一组特征向量.
            x_j: numpy.ndarray, 第二组特征向量.

        Returns:
            核函数映射后的特征向量.
        """
        kappa = self.gamma * np.tanh(self.beta * np.matmul(x_j, x_i.T) + self.theta)

        return np.asmatrix(kappa)


from classicML.api.models.linear_model.linear_discriminant_analysis import LinearDiscriminantAnalysis
from classicML.api.models.linear_model.linear_discriminant_analysis import LinearDiscriminantAnalysis as LDA
from classicML.api.models.linear_model.logistic_regression import LogisticRegression

from classicML.api.models.neural_network.back_propagation_neural_network import BackPropagationNeuralNetwork
from classicML.api.models.neural_network.back_propagation_neural_network import BackPropagationNeuralNetwork as BPNN
from classicML.api.models.neural_network.radial_basis_function_network import RadialBasisFunctionNetwork
from classicML.api.models.neural_network.radial_basis_function_network import RadialBasisFunctionNetwork as RBF

from classicML.api.models.support_vector_machine.support_vector_classifier import SupportVectorClassifier
from classicML.api.models.support_vector_machine.support_vector_classifier import SupportVectorClassifier as SVC

from classicML.api.models.tree.decision_tree_classifier import DecisionTreeClassifier



import numpy as np

from classicML import CLASSICML_LOGGER
from classicML.backend import get_initializer
from classicML.backend import get_optimizer
from classicML.backend import get_loss
from classicML.backend import get_metric


class RadialBasisFunctionNetwork(object):
    """径向基函数网络.

    Attributes:
        seed: int, default=None,
            随机种子.
        initializer: str, 径向基函数网络的初始化器.
        hidden_units: int, 径向基函数网络的隐含层神经元数量.
        optimizer: str, classicML.optimizers.Optimizer 实例, 径向基函数网络的优化器.
        loss: str, classicML.losses.Loss 实例, default='mse'
            径向基函数网络使用的损失函数.
        metric: str, classicML.metrics.Metric 实例, default='accuracy'
            径向基函数网络使用的评估函数.
        parameters: numpy.ndarray,
            径向基函数网络的参数矩阵.
        is_trained: bool, default=None,
            径向基函数网络训练后将被标记为True.
    """
    def __init__(self, seed=None):
        """初始化径向基函数网络.

        Arguments:
            seed: int, default=None,
                随机种子.
        """
        super(RadialBasisFunctionNetwork, self).__init__()
        self.seed = seed

        self.initializer = None
        self.hidden_units = None
        self.optimizer = None
        self.loss = None
        self.metric = None
        self.parameters = None
        self.is_trained = False

    def compile(self, hidden_units, optimizer='rbf', loss='mse', metric='accuracy'):
        """编译径向基函数网络, 配置训练时使用的超参数.

        Arguments:
            hidden_units: int, 径向基函数网络的隐含层神经元数量.
            optimizer: str, classicML.optimizers.Optimizer 实例,
                径向基函数网络使用的优化器.
            loss: str, classicML.losses.Loss 实例, default='mse'
                径向基函数网络使用的损失函数.
            metric: str, classicML.metrics.Metric 实例, default='accuracy'
                径向基函数网络使用的评估函数.

        Notes:
            - 注意RBF只能是用RadialBasisFunctionOptimizer优化器,
              之所以开放优化器选项, 只是为了满足用户修改学习率的需求.
            - 使用交叉熵作为损失函数有潜在异常的可能性,
              除隐含层神经元个数和学习率之外, 建议使用默认参数.
        """
        self.hidden_units = hidden_units

        self.initializer = get_initializer('rbf_normal', self.seed)
        self.optimizer = get_optimizer(optimizer)
        self.loss = get_loss(loss)
        self.metric = get_metric(metric)

    def fit(self, x, y, epochs=1, verbose=True, callbacks=None):
        """训练径向基函数网络.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            verbose: bool, default=True, 可选
                显示日志信息.
            callbacks: list, default=None,
                模型训练过程的中间数据记录器.
        Returns:
            RadialBasisFunctionNetwork实例.
        """
        x = np.asarray(x).astype(float)
        y = np.asarray(y).astype(int)

        # 使用初始化器初始化参数
        self.parameters = self.initializer(hidden_units=self.hidden_units)
        # 使用优化器优化
        self.parameters = self.optimizer(x, y, epochs, self.parameters, verbose, self.loss, self.metric, callbacks)
        # 标记训练完成
        self.is_trained = True

        return self

    def predict(self, x):
        """使用径向基函数网络进行预测.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.

        Returns:
            径向基函数网络预测的概率.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        y_pred, _ = self.optimizer.forward(x, self.parameters)

        return y_pred


import os
from time import time

import numpy as np

from classicML.backend.python.activations import relu
from classicML.backend.python.activations import sigmoid
from classicML.backend.python.activations import softmax

from classicML.backend.python.callbacks import History

from classicML.backend.python.utils import ProgressBar

if os.environ['CLASSICML_ENGINE'] == 'CC':
    from classicML.backend.cc.ops import cc_calculate_error as calculate_error
    from classicML.backend.cc.ops import cc_clip_alpha as clip_alpha
    from classicML.backend.cc.ops import cc_select_second_alpha as select_second_alpha
else:
    from classicML.backend.python.ops import calculate_error
    from classicML.backend.python.ops import clip_alpha
    from classicML.backend.python.ops import select_second_alpha


def _get_optimizer_parameters(args, kwargs):
    """获取优化器的额外参数.

    Arguments:
        args: *args元组.
        kwargs: **kwargs字典.

    Returns:
        额外的参数列表.
    """
    parameters = ['verbose', 'loss', 'metric', 'callbacks']
    for index, arg in enumerate(args):
        parameters[index] = arg

    for kwarg in kwargs:
        for index, parameter in enumerate(parameters):
            if str(kwarg) == parameter:
                parameters[index] = kwargs[kwarg]

    return parameters


def _record_callbacks(callbacks, loss_value, metric_value):
    """记录callbacks数据.

    Arguments:
        callbacks: list, callbacks列表.
        loss_value: float, 当前的损失值.
        metric_value: float, 当前的评估值.
    """
    for callback in callbacks:
        if isinstance(callback, History):
            callback(loss_value, metric_value)
        else:
            pass


class Optimizer(object):
    """优化器的基类.

    Attributes:
        name: str, default=None,
            优化器的名称.
        _progressbar: classicML.backend.python.utils.ProgressBar,
            进度条.
    """
    def __init__(self, name=None):
        """
        Arguments:
            name: str, default=None,
                优化器的名称.
        """
        self.name = name
        self._progressbar = None

    def __call__(self, x, y, epochs, parameters, *args, **kwargs):
        """
        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            parameters: numpy.ndarray, 模型的参数矩阵.
        """
        return self.run(x, y, epochs, parameters, *args, **kwargs)

    def run(self, x, y, epochs, parameters, *args, **kwargs):
        """运行优化器优化参数.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            parameters: numpy.ndarray, 模型的参数矩阵.
        """
        raise NotImplementedError

    def _update_parameters(self, parameters, grad, *args, **kwargs):
        """更新模型的参数.

        Arguments:
            parameters: numpy.ndarray, 模型的参数矩阵.
            grad: numpy.ndarray, 优化器的实时梯度矩阵.
        """
        raise NotImplementedError


class GradientDescent(Optimizer):
    """梯度下降优化器.
    """
    def __init__(self, name='gradient_descent', learning_rate=1e-2):
        """
        Arguments:
            name: str, default=None,
                优化器的名称.
            learning_rate: float, default=1e-2,
                优化器的学习率.
        """
        super(GradientDescent, self).__init__(name=name)
        self.learning_rate = learning_rate

    def run(self, x, y, epochs, beta, *args, **kwargs):
        """运行优化器优化参数.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            beta: numpy.ndarray, 模型的参数矩阵.
            args:
                verbose: bool, 显示日志信息.
                loss: str, classicML.losses.Loss 实例,
                    模型使用的损失函数.
                metric: str, classicML.metrics.Metric 实例,
                    模型使用的评估函数.
        Returns:
            模型的参数矩阵.
        """
        verbose, loss, metric, callbacks = _get_optimizer_parameters(args, kwargs)
        if verbose is True:
            # 实例化进度条
            self._progressbar = ProgressBar(epochs, loss, metric)

        for epoch in range(1, epochs+1):
            current = time()
            # 前向传播
            y_pred, x_hat = self.forward(x, beta)
            # 反向传播
            grad = self.backward(y_pred, y, x_hat)
            # 更新参数
            beta = self._update_parameters(beta, grad)
            # 以下的操作是为了更好可视化的, 不参与实际优化
            if self._progressbar or callbacks is not None:
                if loss.name is 'log_likelihood':
                    loss_value = loss(y, beta, x_hat)
                else:
                    loss_value = loss(y_pred, y)
                metric_value = metric(y_pred, y)

                # 记录callbacks
                if callbacks:
                    _record_callbacks(callbacks, loss_value, metric_value)
                # 进度条显示
                if self._progressbar:
                    self._progressbar(epoch, current, loss_value, metric_value)

        return beta

    def _update_parameters(self, beta, grad, *args, **kwargs):
        """更新模型的参数.

        Arguments:
            beta: numpy.ndarray, 模型的参数矩阵.
            grad: numpy.ndarray, 优化器的实时梯度矩阵.
        """
        beta -= self.learning_rate * grad

        return beta

    @staticmethod
    def forward(x, parameters):
        """优化器前向传播.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            parameters: numpy.ndarray, 模型的参数矩阵.

        Returns:
            预测的标签(概率形式)和参数(x;1)矩阵.
        """
        number_of_sample = x.shape[0]
        parameters = parameters.reshape(-1, 1)

        x_hat = np.c_[x, np.ones((number_of_sample, 1))]
        cache = np.matmul(x_hat, parameters)
        y_pred = sigmoid(cache)

        return y_pred, x_hat

    @staticmethod
    def backward(y_pred, y_true, x_hat):
        """优化器反向传播.

        Arguments:
            y_pred: numpy.ndarray, 预测的标签(概率形式).
            y_true: numpy.ndarray, 真实的标签.
            x_hat: numpy.ndarray, 属性的参数矩阵.

        Returns:
            优化器的实时梯度矩阵.
        """
        y_true = y_true.reshape(-1, 1)
        error = y_true - y_pred
        grad = np.sum((-x_hat * error), axis=0)
        grad = grad.reshape(-1, 1)

        return grad


class NewtonMethod(Optimizer):
    """牛顿法优化器.
    """
    def __init__(self, name='newton_method'):
        """
        Arguments:
            name: str, default=None,
                优化器的名称.
        """
        super(NewtonMethod, self).__init__(name=name)

    def run(self, x, y, epochs, beta, *args, **kwargs):
        """运行优化器优化参数.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            beta: numpy.ndarray, 模型的参数矩阵.
            args:
                verbose: bool, 显示日志信息.
                loss: str, classicML.losses.Loss 实例,
                    模型使用的损失函数.
                metric: str, classicML.metrics.Metric 实例,
                    模型使用的评估函数.
        Returns:
            模型的参数矩阵.
        """
        verbose, loss, metric, callbacks = _get_optimizer_parameters(args, kwargs)
        if verbose:
            # 实例化进度条
            self._progressbar = ProgressBar(epochs, loss, metric)

        for epoch in range(1, epochs + 1):
            current = time()
            # 前向传播
            y_pred, x_hat = self.forward(x, beta)
            # 反向传播
            grad = self.backward(y_pred, y, x_hat)
            # 求解海森矩阵
            hessian = self._get_hessian_matrix(y_pred, x_hat)
            # 更新参数
            beta = self._update_parameters(beta, grad, hessian)
            # 以下的操作是为了更好可视化的, 不参与实际优化
            if self._progressbar or callbacks is not None:
                if loss.name is 'log_likelihood':
                    loss_value = loss(y, beta, x_hat)
                else:
                    loss_value = loss(y_pred, y)
                metric_value = metric(y_pred, y)

                # 进度条显示
                if self._progressbar:
                    self._progressbar(epoch, current, loss_value, metric_value)
                # 记录callbacks
                if callbacks:
                    _record_callbacks(callbacks, loss_value, metric_value)

        return beta

    def _update_parameters(self, beta, grad, *args, **kwargs):
        """更新模型的参数.

        Arguments:
            beta: numpy.ndarray, 模型的参数矩阵.
            grad: numpy.ndarray, 优化器的实时梯度矩阵.
            args:
                hessian: numpy.ndarray, 海森矩阵.
        """
        hessian = args[0]
        beta -= np.matmul(np.linalg.inv(hessian), grad)

        return beta

    @staticmethod
    def forward(x, parameters):
        """优化器前向传播.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            parameters: numpy.ndarray, 模型的参数矩阵.

        Returns:
            预测的标签(概率形式)和参数(x;1)矩阵.
        """
        number_of_sample = x.shape[0]
        parameters = parameters.reshape(-1, 1)

        x_hat = np.c_[x, np.ones((number_of_sample, 1))]
        cache = np.matmul(x_hat, parameters)
        y_pred = sigmoid(cache)

        return y_pred, x_hat

    @staticmethod
    def backward(y_pred, y_true, x_hat):
        """优化器反向传播.

        Arguments:
            y_pred: numpy.ndarray, 预测的标签(概率形式).
            y_true: numpy.ndarray, 真实的标签.
            x_hat: numpy.ndarray, 属性的参数矩阵.

        Returns:
            优化器的实时梯度矩阵.
        """
        y_true = y_true.reshape(-1, 1)
        error = y_true - y_pred
        grad = np.sum((-x_hat * error), axis=0)
        grad = grad.reshape(-1, 1)

        return grad

    @staticmethod
    def _get_hessian_matrix(y_pred, x_hat):
        """计算海森矩阵(二阶导数).

        Arguments:
            y_pred: numpy.ndarray, 预测的标签(概率形式).
            x_hat: numpy.ndarray, 属性的参数矩阵.

        Returns:
            海森矩阵.
        """
        P = np.eye(x_hat.shape[0]) * y_pred * (1 - y_pred)
        H = np.matmul(np.matmul(x_hat.T, P), x_hat)

        return H


class StochasticGradientDescent(Optimizer):
    """随机梯度下降优化器.

    Notes:
        - 如果想固定随机种子实现, 实现复现的话,
          请在模型实例化的时候将随机种子置为一个常整数.
    """
    def __init__(self, name='stochastic_gradient_descent', learning_rate=1e-2):
        """
        Arguments:
            name: str, default=None,
                优化器的名称.
            learning_rate: float, default=1e-2,
                优化器的学习率.
        """
        super(StochasticGradientDescent, self).__init__(name=name)
        self.learning_rate = learning_rate

    def run(self, x, y, epochs, parameters, *args, **kwargs):
        """运行优化器优化参数.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            parameters: numpy.ndarray, 模型的参数矩阵.
            args:
                verbose: bool, 显示日志信息.
                loss: str, classicML.losses.Loss 实例,
                    模型使用的损失函数.
                metric: str, classicML.metrics.Metric 实例,
                    模型使用的评估函数.
                callbacks: list,
                    模型训练过程的中间数据记录器.
        Returns:
            模型的参数矩阵.
        """
        verbose, loss, metric, callbacks = _get_optimizer_parameters(args, kwargs)
        num_of_features = x.shape[0]

        if verbose is True:
            # 实例化进度条
            self._progressbar = ProgressBar(epochs, loss, metric)

        for epoch in range(1, epochs + 1):
            current = time()
            # 随机选取样本
            random_index = np.random.randint(0, num_of_features)
            # 前向传播
            y_pred, caches = self.forward(x[[random_index], :], parameters)
            # 反向传播
            # Notes: 此时y_true是个scalar
            grad = self.backward(y_pred, np.asarray([y[random_index]]), caches)
            # 更新参数
            parameters = self._update_parameters(parameters, grad)
            # 以下的操作是为了更好可视化的, 不参与实际优化
            if self._progressbar or callbacks is not None:
                # 由于每次只更新一个样本导致预测值的形状和标签不一致,
                # 因此需要多做一次前向传播
                y_pred, _ = self.forward(x, parameters)

                loss_value = loss(y_pred, y)
                metric_value = metric(y_pred, y)

                # 进度条显示
                if self._progressbar:
                    self._progressbar(epoch, current, loss_value, metric_value)
                # 记录callbacks
                if callbacks:
                    _record_callbacks(callbacks, loss_value, metric_value)

        return parameters

    def _update_parameters(self, parameters, grad, *args, **kwargs):
        """更新模型的参数.

        Arguments:
            parameters: numpy.ndarray, 模型的参数矩阵.
            grad: numpy.ndarray, 优化器的实时梯度矩阵.

        Returns:
            模型的参数矩阵.
        """
        num_of_matrix_ops = len(parameters) // 2

        for i in range(1, num_of_matrix_ops + 1):
            parameters['w' + str(i)] -= self.learning_rate * grad['dw' + str(i)]
            parameters['b' + str(i)] -= self.learning_rate * grad['db' + str(i)]

        return parameters

    @staticmethod
    def forward(x, parameters):
        """优化器前向传播.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            parameters: numpy.ndarray, 模型的参数矩阵.

        Returns:
            预测的标签(概率形式)和参数矩阵缓存.
        """
        num_of_matrix_ops = len(parameters) // 2
        caches = []

        a = x  # 除了第一个输入层以外, 其他的是经过激活后的

        for i in range(1, num_of_matrix_ops):
            # 提取参数
            w = parameters['w' + str(i)]
            b = parameters['b' + str(i)]
            x = a  # 提取上一轮的a作为输入

            # 计算输出并激活
            z = np.matmul(x, w.T) + b
            a = relu(z)
            # 保存参数缓存
            cache = (x, w, a)
            caches.append(cache)

        # 输出层的激活函数不同, 单独处理
        w_output = parameters['w' + str(num_of_matrix_ops)]
        b_output = parameters['b' + str(num_of_matrix_ops)]

        z = np.matmul(a, w_output.T) + b_output
        if w_output.shape[0] == 1:
            y_pred = sigmoid(z)
        else:
            y_pred = softmax(z)

        cache = (a, w_output, y_pred)
        caches.append(cache)

        return y_pred, caches

    @staticmethod
    def backward(y_pred, y_true, caches):
        """优化器反向传播.

        Arguments:
            y_pred: numpy.ndarray, 预测的标签(概率形式).
            y_true: numpy.ndarray, 真实的标签.
            caches: numpy.ndarray, 参数缓存.

        Returns:
            优化器的实时梯度矩阵字典.
        """
        num_of_caches = len(caches)
        grad = {}

        if y_true.ndim == 1:
            y_true = y_true.reshape(-1, 1)

        # 提取缓存
        (x, w, a) = caches[num_of_caches - 1]
        if y_true.shape[1] == 1:
            # 在二分类的情况下, y_true的标签为零的时候会导致无法计算梯度
            # 这里为了解决这个问题进行了归一化, (-无穷, 0)和(0, 无穷),
            # 实际上到不了无穷, 在10以内
            y_true = -(y_true / y_pred - (1 - y_true) / (1 - y_pred))
            da = sigmoid.diff(y_pred, a, y_true)
        else:
            da = softmax.diff(y_true, a)

        # 计算梯度并保存
        dw, db, da_ = StochasticGradientDescent._get_grad(da, y_true, caches[num_of_caches - 1])
        grad['dw' + str(num_of_caches)] = dw
        grad['db' + str(num_of_caches)] = db
        grad['da' + str(num_of_caches)] = da_

        for cache in range(num_of_caches - 1, 0, -1):
            (x, w, a) = caches[cache - 1]

            output = grad.get('da' + str(cache + 1))
            da = relu.diff(output, a)

            # 计算梯度并保存
            dw, db, da_ = StochasticGradientDescent._get_grad(da, y_true, caches[cache - 1])
            grad['dw' + str(cache)] = dw
            grad['db' + str(cache)] = db
            grad['da' + str(cache)] = da_

        return grad

    @staticmethod
    def _get_grad(da, y_true, cache):
        """获取梯度.

        Arguments:
            da: numpy.ndarray, 输出张量的梯度.
            y_true: numpy.ndarray, 真实的标签.
            cache: numpy.ndarray, 参数缓存.

        Returns:
            参数矩阵, 偏置矩阵, 输出张量的梯度.
        """

        # 提取缓存
        (x, w, a) = cache

        dw = np.matmul(da.T, x) / y_true.shape[0]
        db = np.sum(da, axis=0, keepdims=True) / y_true.shape[0]
        da_ = np.matmul(da, w)

        return dw, db, da_


class Adam(StochasticGradientDescent):
    """自适应矩估计优化器.

    References:
        - [Kingma et al., 2014](http://arxiv.org/abs/1412.6980)
            参照算法1具体实现.

    Notes:
        - 如果想固定随机种子实现, 实现复现的话,
          请在模型实例化的时候将随机种子置为一个常整数.
        - 并采用随机梯度下降作为基础优化算法实现的.
        - 超参数epsilon按照其他的机器学习框架设置的1e-7,
          非原论文的1e-8.
    """
    def __init__(self,
                 name='adam',
                 learning_rate=1e-3,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=1e-7):
        """
        Arguments:
            name: str, default=None,
                优化器的名称.
            learning_rate: float, default=1e-3,
                优化器的学习率.
            beta_1: float, default=0.9,
                一阶矩估计衰减率.
            beta_2: float, default=0.999
                二阶矩估计衰减率.
            epsilon: float, default=1e-7,
                数值稳定的小常数.
        """
        super(Adam, self).__init__(name=name)
        self.learning_rate = learning_rate
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon

    def run(self, x, y, epochs, parameters, *args, **kwargs):
        """运行优化器优化参数.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            parameters: numpy.ndarray, 模型的参数矩阵.
            args:
                verbose: bool, 显示日志信息.
                loss: str, classicML.losses.Loss 实例,
                    模型使用的损失函数.
                metric: str, classicML.metrics.Metric 实例,
                    模型使用的评估函数.
                callbacks: list,
                    模型训练过程的中间数据记录器.
        Returns:
            模型的参数矩阵.
        """
        verbose, loss, metric, callbacks = _get_optimizer_parameters(args, kwargs)
        num_of_features = x.shape[0]

        # 对Adam进行初始化
        m, v = self._initializer(parameters)

        if verbose is True:
            # 实例化进度条
            self._progressbar = ProgressBar(epochs, loss, metric)

        for epoch in range(1, epochs + 1):
            current = time()
            # 随机选取样本
            random_index = np.random.randint(0, num_of_features)
            # 前向传播
            y_pred, caches = self.forward(x[[random_index], :], parameters)
            # 反向传播
            # Notes: 此时y_true是个scalar
            grad = self.backward(y_pred, np.asarray([y[random_index]]), caches)
            # 更新参数
            parameters, m, v = self._update_parameters(parameters, grad, m, v, epoch)
            # 以下的操作是为了更好可视化的, 不参与实际优化
            if self._progressbar or callbacks is not None:
                # 由于每次只更新一个样本导致预测值的形状和标签不一致,
                # 因此需要多做一次前向传播
                y_pred, _ = self.forward(x, parameters)

                loss_value = loss(y_pred, y)
                metric_value = metric(y_pred, y)

                # 进度条显示
                if self._progressbar:
                    self._progressbar(epoch, current, loss_value, metric_value)
                # 记录callbacks
                if callbacks:
                    _record_callbacks(callbacks, loss_value, metric_value)

        return parameters

    def _update_parameters(self, parameters, grad, *args, **kwargs):
        """更新模型的参数.

        Arguments:
            parameters: numpy.ndarray, 模型的参数矩阵.
            grad: numpy.ndarray, 优化器的实时梯度矩阵.
            args:
                moment_vector_1: dict of numpy.ndarray,
                    一阶向量.
                moment_vector_2: dict of numpy.ndarray,
                    二阶向量.
                epoch: int, 当前的训练轮数.

        Returns:
            模型的参数矩阵.
        """
        m, v, epoch = args[0], args[1], args[2]

        num_of_matrix_ops = len(parameters) // 2

        for i in range(1, num_of_matrix_ops + 1):
            # 更新一阶矩估计误差
            m['dw' + str(i)] = self.beta_1 * m['dw' + str(i)] + (1 - self.beta_1) * grad['dw' + str(i)]
            m['db' + str(i)] = self.beta_1 * m['db' + str(i)] + (1 - self.beta_1) * grad['db' + str(i)]
            # 更新二阶矩估计误差
            v['dw' + str(i)] = self.beta_2 * v['dw' + str(i)] + (1 - self.beta_2) * (grad['dw' + str(i)] ** 2)
            v['db' + str(i)] = self.beta_2 * v['db' + str(i)] + (1 - self.beta_2) * (grad['db' + str(i)] ** 2)
            # 矫正一阶矩估计误差
            m_hat_w = m['dw' + str(i)] / (1 - np.power(self.beta_1, epoch))
            m_hat_b = m['db' + str(i)] / (1 - np.power(self.beta_1, epoch))
            # 矫正二阶矩估计误差
            v_hat_w = v['dw' + str(i)] / (1 - np.power(self.beta_2, epoch))
            v_hat_b = v['db' + str(i)] / (1 - np.power(self.beta_2, epoch))
            # 更新参数
            parameters['w' + str(i)] -= self.learning_rate * m_hat_w / (np.sqrt(v_hat_w) + self.epsilon)
            parameters['b' + str(i)] -= self.learning_rate * m_hat_b / (np.sqrt(v_hat_b) + self.epsilon)

        return parameters, m, v
    
    @staticmethod
    def _initializer(parameters):
        """初始化一阶和二阶向量, 算法一实现全初始化为零.

        Arguments:
            parameters: numpy.ndarray, 模型的参数矩阵.

        Returns:
            一阶和二阶向量字典.
        """
        num_of_matrix_ops = len(parameters) // 2

        moment_vector_1 = {}
        moment_vector_2 = {}

        for i in range(1, num_of_matrix_ops + 1):
            moment_vector_1['dw' + str(i)] = np.zeros(parameters['w' + str(i)].shape)
            moment_vector_1['db' + str(i)] = np.zeros(parameters['b' + str(i)].shape)

            moment_vector_2['dw' + str(i)] = np.zeros(parameters['w' + str(i)].shape)
            moment_vector_2['db' + str(i)] = np.zeros(parameters['b' + str(i)].shape)

        return moment_vector_1, moment_vector_2


class RadialBasisFunctionOptimizer(Optimizer):
    """径向基函数优化器.
    """
    def __init__(self, name='rbf', learning_rate=1e-2):
        """
        Arguments:
            name: str, default=None,
                优化器的名称.
            learning_rate: float, default=1e-2,
                优化器的学习率.
        """
        super(RadialBasisFunctionOptimizer, self).__init__(name=name)
        self.learning_rate = learning_rate

    def run(self, x, y, epochs, parameters, *args, **kwargs):
        """运行优化器优化参数.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            parameters: numpy.ndarray, 模型的参数矩阵.
            args:
                verbose: bool, 显示日志信息.
                loss: str, classicML.losses.Loss 实例,
                    模型使用的损失函数.
                metric: str, classicML.metrics.Metric 实例,
                    模型使用的评估函数.
                callbacks: list,
                    模型训练过程的中间数据记录器.
        Returns:
            模型的参数矩阵.
        """
        verbose, loss, metric, callbacks = _get_optimizer_parameters(args, kwargs)

        if verbose is True:
            # 实例化进度条
            self._progressbar = ProgressBar(epochs, loss, metric)

        for epoch in range(1, epochs+1):
            current = time()
            # 前向传播
            y_pred, cache = self.forward(x, parameters)
            # 反向传播
            grad = self.backward(y_pred, y, cache)
            # 更新参数
            parameters = self._update_parameters(parameters, grad)
            # 以下的操作是为了更好可视化的, 不参与实际优化
            if self._progressbar or callbacks is not None:
                loss_value = loss(y_pred, y)
                metric_value = metric(y_pred, y)
                # 进度条显示
                if self._progressbar:
                    self._progressbar(epoch, current, loss_value, metric_value)
                # 记录callbacks
                if callbacks:
                    _record_callbacks(callbacks, loss_value, metric_value)

        return parameters

    def _update_parameters(self, parameters, grad, *args, **kwargs):
        """更新模型的参数.

        Arguments:
            parameters: numpy.ndarray, 模型的参数矩阵.
            grad: numpy.ndarray, 优化器的实时梯度矩阵.

        Returns:
            模型的参数矩阵.
        """
        parameters['w'] -= self.learning_rate * grad['dw']
        parameters['b'] -= self.learning_rate * grad['db']
        parameters['beta'] -= self.learning_rate * grad['dbeta']

        return parameters

    @staticmethod
    def forward(x, parameters):
        """优化器前向传播.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            parameters: numpy.ndarray, 模型的参数矩阵.

        Returns:
            预测的标签(概率形式)和参数矩阵缓存.
        """
        number_of_sample = x.shape[0]
        # 提取参数
        w, b, c, beta = parameters['w'], parameters['b'], parameters['c'], parameters['beta']
        # 获取隐含层神经元的数量
        units = c.shape[0]
        # 初始化高斯径向基函数
        rho = np.zeros((number_of_sample, units))
        x_ci = np.zeros((number_of_sample, units))
        for unit in range(units):
            x_ci[:, unit] = np.linalg.norm(x - c[[unit], ], axis=1) ** 2
            rho[:, unit] = np.exp(-beta[0, unit] * x_ci[:, unit])

        y_pred = np.matmul(rho, w.T) + b
        cache = (rho, x_ci, w, beta)

        return y_pred, cache

    @staticmethod
    def backward(y_pred, y_true, cache):
        """优化器反向传播.

        Arguments:
            y_pred: numpy.ndarray, 预测的标签(概率形式).
            y_true: numpy.ndarray, 真实的标签.
            cache: numpy.ndarray, 参数缓存.

        Returns:
            优化器的实时梯度矩阵.
        """
        (rho, x_ci, w, beta) = cache
        grad = {}

        dy = y_pred - y_true  # 这里虽然叫做dy, 但其实是真实值和预测值之间的误差
        dw = np.matmul(dy.T, rho) / y_true.shape[0]
        db = np.sum(dy, axis=0, keepdims=True) / y_true.shape[0]
        drho = np.matmul(y_pred, w)
        dbeta = np.sum(drho * rho * (-x_ci), axis=0, keepdims=True) / y_true.shape[0]

        grad['dw'] = dw
        grad['db'] = db
        grad['dbeta'] = dbeta

        return grad


class SequentialMinimalOptimization(Optimizer):
    """序列最小最优化算法. SMO算法是一种启发式算法,
    即每次优化两个变量, 使之满足KKT条件; 不断迭代, 最后使得全部变量满足KKT条件.
    整个SMO算法包括: 求解两个变量的二次规划问题和选择变量的启发式方法.

    Attributes:
        alphas: numpy.ndarray,
            拉格朗日乘子数组.
        non_bound_alphas: numpy.ndarray,
            非边界拉格朗日乘子.(硬间隔下和非零拉格朗日乘子一样, 软间隔是非零拉格朗日乘子的子集.)
        error_cache: numpy.ndarray,
            KKT条件的违背值缓存.
        non_zero_alphas: numpy.ndarray,
            非零拉格朗日乘子.
        b: float, default=0,
            偏置项.
        C: float, default=None
            软间隔正则化系数.
        kernel: str, classicML.kernel.Kernels 实例, default=None
            分类器使用的核函数.
        tol: float, default=None
            停止训练的最大误差值.
        epochs: int, default=None
            最大的训练轮数, 如果是-1则表示需要所有的样本满足条件时,
            才能停止训练, 即没有限制.

    References:
        - [李航, 2012.] 统计学习方法 P124~P131
    """
    def __init__(self, name='SMO'):
        super(SequentialMinimalOptimization, self).__init__(name=name)

        self.alphas = None
        self.non_bound_alphas = None
        self.error_cache = None
        self.non_zero_alphas = None
        self.b = None

        self.C = None
        self.kernel = None
        self.tol = None
        self.epochs = None

    def run(self, x, y, *args, **kwargs):
        """运行优化器优化参数.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            args:
                C: float,
                    软间隔正则化系数.
                kernel: str, classicML.kernel.Kernels 实例,
                    分类器使用的核函数.
                tol: float,
                    停止训练的最大误差值.
                epochs: int,
                    最大的训练轮数, 如果是-1则表示需要所有的样本满足条件时,
                    才能停止训练, 即没有限制.
        Returns:
            分类器的支持向量下标数组, 支持向量数组, 拉格朗日乘子数组, 支持向量对应的标签数组和偏置项.
        """
        self.C, self.kernel, self.tol, self.epochs = args[0], args[1], args[2], args[3]

        epoch = 0
        number_of_sample = x.shape[0]
        entire_flag = True

        # 初始化为numpy.ndarray数组, 在Python下更新后会自动转换数据类型为numpy.ndarray,
        # 但是, CC在第一次输入的时候为int, 这里强制为numpy.ndarray用以解决CC下不能动态类型输入.
        self.b = np.asarray([0])
        # 初始化返回参数, 在未知样本的数量时无法初始化.
        self.alphas = np.zeros((number_of_sample, ))
        self.non_bound_alphas = np.zeros((number_of_sample, ))
        self.error_cache = np.zeros((number_of_sample, ))
        self.non_zero_alphas = np.zeros((number_of_sample, ), dtype=bool)

        while (self.epochs == -1) or (epoch < self.epochs):
            pair_of_alpha_changed = 0
            epoch += 1  # 更新迭代次数, 在代码块结尾更新有可能因为提前返回而计数不准确.

            if entire_flag:  # 第一次必须全部全部遍历一遍, 因为初始值全部为零.
                for sample in range(number_of_sample):
                    pair_of_alpha_changed += self._update_parameters(x, y, sample, number_of_sample)
            else:
                non_bound_index = self.non_bound_alphas.nonzero()[0]
                for sample in non_bound_index:
                    pair_of_alpha_changed += self._update_parameters(x, y, sample, number_of_sample)

            if entire_flag is True:
                # 遍历所有样本还是没有更新, 就退出循环.
                if pair_of_alpha_changed == 0:
                    support = self.non_zero_alphas.nonzero()[0]
                    support_vector = x[support]
                    support_alpha = self.alphas[self.non_zero_alphas]
                    support_y = y[support]

                    return support, support_vector, support_alpha, support_y, self.b
                entire_flag = False
            elif pair_of_alpha_changed == 0:
                entire_flag = True

        support = self.non_zero_alphas.nonzero()[0]
        support_vector = x[support]
        support_alpha = self.alphas[self.non_zero_alphas]
        support_y = y[support]

        return support, support_vector, support_alpha, support_y, self.b

    def _update_parameters(self, x, y, *args, **kwargs):
        """SMO算法的内循环(更新参数的具体实现), 寻找第二个要更新的变量alpha_j, 并进行更新.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            args:
                i: int, 第i个样本.
                number_of_sample: int, 样本的总数.

        Returns:
            更新是否成功的标记.
        """
        i, number_of_sample = args[0], args[1]
        y_i = y[i, :]
        alpha_i = self.alphas[i]

        # 提取违背值, 保存为缓存(可能要实时计算).
        if self.non_bound_alphas[i]:
            error_i = self.error_cache[i]
        else:
            error_i = calculate_error(x, y, i, self.kernel, self.alphas, self.non_zero_alphas, self.b)

        # 更新的变化量的绝对值要大于tol, 且alpha要满足软间隔C的条件限制.
        # TODO(Steve R. Sun): 直接使用绝对值会导致异常, 但是这样就能正常运行.
        if ((y_i * error_i < -self.tol) and (alpha_i < self.C)) or ((y_i * error_i > self.tol) and (0 < alpha_i)):
            # 存在非边界拉格朗日乘子.
            if np.sum(self.non_bound_alphas) > 0:
                j, error_j = select_second_alpha(error_i, self.error_cache, self.non_bound_alphas)
                if self._update_alpha(x, y, i, j, error_i, error_j):
                    return 1

            # 试图逐元素强制更新.
            for j in np.random.permutation(number_of_sample):
                error_j = calculate_error(x, y, j, self.kernel, self.alphas, self.non_zero_alphas, self.b)
                if self._update_alpha(x, y, i, j, error_i, error_j):
                    return 1

        return 0

    def _update_alpha(self, x, y, i, j, error_i, error_j):
        """更新拉格朗日乘子.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            i: int, 第i个样本.
            j: int, 第j个样本.
            error_i: float, 第i个样本的违背值.
            error_j: float, 第j个样本的违背值.

        Returns:
            是否更新成功.
        """
        if i == j:
            return False

        alpha_i_old = self.alphas[i].copy()
        alpha_j_old = self.alphas[j].copy()
        y_i = y[i, 0]
        y_j = y[j, 0]

        if y_i != y_j:
            low = max(0.0, alpha_j_old - alpha_i_old)
            high = min(self.C, self.C + alpha_j_old - alpha_i_old)
        else:
            low = max(0.0, alpha_i_old + alpha_j_old - self.C)
            high = min(self.C, alpha_i_old + alpha_j_old)

        if low == high:  # 二维空间对角线重合.
            return False

        x_i = x[[i], :]
        x_j = x[[j], :]

        kappa_ii = self.kernel(x_i, x_i)
        kappa_ij = self.kernel(x_i, x_j)
        kappa_ji = self.kernel(x_j, x_i)
        kappa_jj = self.kernel(x_j, x_j)

        eta = kappa_ii + kappa_jj - 2 * kappa_ij
        if eta <= 0:  # 2-范数小于零.
            return False

        alpha_j_new = alpha_j_old + y_j * (error_i - error_j) / eta
        alpha_j_new = clip_alpha(alpha_j_new, low, high)

        if np.abs(alpha_j_new - alpha_j_old) < 1e-5:  # 更新幅度过小.
            return False

        alpha_i_new = alpha_i_old + y_i * y_j * (alpha_j_old - alpha_j_new)

        # 两个变量优化后要重新计算偏置.
        b_i = (-error_i - y_i * kappa_ii * (alpha_i_new - alpha_i_old)
               - y_j * kappa_ji * (alpha_j_new - alpha_j_old) + self.b)
        b_j = (-error_j - y_i * kappa_ij * (alpha_i_new - alpha_i_old)
               - y_j * kappa_jj * (alpha_j_new - alpha_j_old) + self.b)
        self.b = (b_i + b_j) / 2  # 这里采用的周志华机器学习中的方法, 求解平均值更为鲁棒.

        self.alphas[i] = alpha_i_new
        self.alphas[j] = alpha_j_new

        # 更新缓存.
        self._update_error_cache(x, y)
        self._update_alpha_cache(alpha_i_new, i)
        self._update_alpha_cache(alpha_j_new, j)

        return True

    def _update_error_cache(self, x, y):
        """更新违背值缓存.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
        """
        for i in self.non_bound_alphas.nonzero()[0]:
            self.error_cache[i] = calculate_error(x, y, i, self.kernel, self.alphas, self.non_zero_alphas, self.b)

    def _update_alpha_cache(self, alpha, index):
        """更新拉格朗日乘子缓存.

        Arguments:
            alpha: float, 拉格朗日乘子.
            index: int, 拉格朗日乘子的下标.
        """
        self.non_zero_alphas[index] = (alpha > 0)
        self.non_bound_alphas[index] = int(0 < alpha < self.C)


# Aliases.
SGD = StochasticGradientDescent
SMO = SequentialMinimalOptimization


from glob import glob
from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext

with open('README.md', 'r') as fh:
    long_description = fh.read()

extension_modules = [
    Pybind11Extension(
        'classicML/backend/cc/ops',
        sorted(glob('classicML/backend/cc/*.cc')),
        include_dirs=[
            '/usr/local/include/eigen3',  # /path/to/eigen3/download 默认为macOS的路径, Ubuntu是/usr/include/eigen3
        ],
        language='c++',
    )
]

setup(
    name='classicML',
    version='0.5b1',
    author='Steve R. Sun',
    license='Apache Software License',
    author_email='s1638650145@gmail.com',
    description='An easy-to-use ML framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sun1638650145/classicML',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.19.2',
        'pandas>=1.1.3',
        'matplotlib>=3.3.2',
        'psutil>=5.7.2',
    ],
    cmdclass={'build_ext': build_ext},
    ext_modules=extension_modules,
)



import numpy as np

from classicML import CLASSICML_LOGGER
from classicML.backend import get_initializer
from classicML.backend import get_optimizer
from classicML.backend import get_loss
from classicML.backend import get_metric


class LogisticRegression(object):
    """逻辑回归.

    Attributes:
        seed: int, default=None,
            随机种子.
        initializer: str or classicML.initializers.Initializer 实例, default='random_normal'
            初始化器.
        optimizer: str, classicML.optimizers.Optimizer 实例, default='newton'
            模型使用的优化器.
        loss: str, classicML.losses.Loss 实例, default='log_likelihood'
            模型使用的损失函数.
        metric: str, classicML.metrics.Metric 实例, default='accuracy'
            模型使用的评估函数.
        beta: numpy.ndarray, 模型的参数矩阵.
        is_trained: bool, default=None, 模型训练后将被标记为True.
    """
    def __init__(self, seed=None, initializer=None):
        """初始化模型.

        Arguments:
             seed: int, default=None,
                随机种子.
             initializer: str or classicML.initializers.Initializer 实例, default='random_normal'
                初始化器.
        """
        super(LogisticRegression, self).__init__()
        self.seed = seed
        self.initializer = initializer

        self.initializer = None
        self.optimizer = None
        self.loss = None
        self.metric = None
        self.beta = None
        self.is_trained = False

    def compile(self, optimizer='newton', loss='log_likelihood', metric='accuracy'):
        """编译模型, 配置训练时使用的超参数.

        Arguments:
            optimizer: str, classicML.optimizers.Optimizer 实例, default='newton'
                模型使用的优化器.
            loss: str, classicML.losses.Loss 实例, default='log_likelihood'
                模型使用的损失函数.
            metric: str, classicML.metrics.Metric 实例, default='accuracy'
                模型使用的评估函数.
        """
        self.initializer = get_initializer(self.initializer, self.seed)
        self.optimizer = get_optimizer(optimizer)
        self.loss = get_loss(loss)
        self.metric = get_metric(metric)

    def fit(self, x, y, epochs=1, verbose=True, callbacks=None):
        """训练模型.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1, 训练的轮数.
            verbose: bool, default=True, 可选
                显示日志信息.
            callbacks: list, default=None,
                模型训练过程的中间数据记录器.
        Returns:
            LogisticRegression实例.
        """
        x = np.asarray(x).astype(float)
        y = np.asarray(y).astype(int)

        _attributes_of_feature = x.shape[1]

        # 使用初始化器初始化参数
        self.beta = self.initializer(attributes_or_structure=_attributes_of_feature)
        # 使用优化器优化
        self.beta = self.optimizer(x, y, epochs, self.beta, verbose, self.loss, self.metric, callbacks)
        # 标记训练完成
        self.is_trained = True

        return self

    def predict(self, x):
        """模型进行预测.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.

        Returns:
            LogisticRegression预测的概率.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        y_pred, _ = self.optimizer.forward(x, self.beta)
        y_pred = np.squeeze(y_pred)

        return y_pred


import numpy as np


class Initializer(object):
    """初始化器的基类.

    Attributes:
        name: str, default=None,
            初始化器的名称.
        seed: int, default=None,
            初始化器的随机种子.

    Raises:
       NotImplementedError: __call__方法需要用户实现.
    """
    def __init__(self, name=None, seed=None):
        """
        Arguments:
            name: str, default=None,
                初始化器的名称.
            seed: int, default=None,
                初始化器的随机种子.
        """
        self.name = name
        self.seed = seed
        np.random.seed(self.seed)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class RandomNormal(Initializer):
    """正态分布随机初始化器.
    """
    def __init__(self, name='random_normal', seed=None):
        super(RandomNormal, self).__init__(name=name, seed=seed)

    def __call__(self, attributes_or_structure):
        """
        Arguments:
            attributes_or_structure: int or list,
                如果是逻辑回归就是样本的特征数,
                如果是神经网络, 就是定义神经网络的网络结构.
        """
        if isinstance(attributes_or_structure, int):
            parameters = np.random.randn(attributes_or_structure + 1, 1)  # 初始化属性数+1(偏置项b)
        else:
            parameters = {}
            num_of_layers = len(attributes_or_structure)

            for layer in range(num_of_layers - 1):
                w = np.random.randn(attributes_or_structure[layer + 1], attributes_or_structure[layer])
                b = np.zeros((1, attributes_or_structure[layer + 1]))
                parameters['w' + str(layer + 1)] = w
                parameters['b' + str(layer + 1)] = b

        return parameters


class HeNormal(Initializer):
    """He正态分布随机初始化器.

    References:
        - [He et al., 2015](https://www.cv-foundation.org/openaccess/content_iccv_2015/html/He_Delving_Deep_into_ICCV_2015_paper.html)
          ([pdf](https://www.cv-foundation.org/openaccess/content_iccv_2015/papers/He_Delving_Deep_into_ICCV_2015_paper.pdf))
    """
    def __init__(self, name='he_normal', seed=None):
        super(HeNormal, self).__init__(name=name, seed=seed)

    def __call__(self, attributes_or_structure):
        """初始化方式为W~N(0, sqrt(2/N_in)), 其中N_in为对应连接的输入层的神经元个数.

        Arguments:
            attributes_or_structure: int or list,
                如果是逻辑回归就是样本的特征数,
                如果是神经网络, 就是定义神经网络的网络结构.
        """
        if isinstance(attributes_or_structure, int):
            parameters = (np.random.randn(attributes_or_structure + 1, 1)
                          * np.sqrt(2 / attributes_or_structure))  # 初始化属性数+1(偏置项b)
        else:
            parameters = {}
            num_of_layers = len(attributes_or_structure)

            for layer in range(num_of_layers - 1):
                w = (np.random.randn(attributes_or_structure[layer + 1], attributes_or_structure[layer])
                     * np.sqrt(2 / attributes_or_structure[layer]))
                b = np.zeros((1, attributes_or_structure[layer + 1]))
                parameters['w' + str(layer + 1)] = w
                parameters['b' + str(layer + 1)] = b

        return parameters


class XavierNormal(Initializer):
    """Xavier正态分布随机初始化器,
        也叫做Glorot正态分布随机初始化器.

    References:
        - [Glorot et al., 2010](http://proceedings.mlr.press/v9/glorot10a.html)
          ([pdf](http://jmlr.org/proceedings/papers/v9/glorot10a/glorot10a.pdf))
    """
    def __init__(self, name='xavier_normal', seed=None):
        super(XavierNormal, self).__init__(name=name, seed=seed)

    def __call__(self, attributes_or_structure):
        """初始化方式为W~N(0, sqrt(2/N_in+N_out)),
            其中N_in为对应连接的输入层的神经元个数, N_out为本层的神经元个数.

        Arguments:
            attributes_or_structure: int or list,
                如果是逻辑回归就是样本的特征数,
                如果是神经网络, 就是定义神经网络的网络结构.
        """
        if isinstance(attributes_or_structure, int):
            # 逻辑回归没有多层结构
            # 令 2 / (N_in + N_out) = 2 / N_in * 2, 即 N_in
            parameters = (np.random.randn(attributes_or_structure + 1, 1)
                          * np.sqrt(attributes_or_structure))
        else:
            parameters = {}
            num_of_layers = len(attributes_or_structure)

            for layer in range(num_of_layers - 1):
                w = (np.random.randn(attributes_or_structure[layer + 1], attributes_or_structure[layer])
                     * np.sqrt(2 / (attributes_or_structure[layer] + attributes_or_structure[layer + 1])))
                b = np.zeros((1, attributes_or_structure[layer + 1]))
                parameters['w' + str(layer + 1)] = w
                parameters['b' + str(layer + 1)] = b

        return parameters


class GlorotNormal(XavierNormal):
    """Glorot正态分布随机初始化器.
        具体实现参看XavierNormal.
    """
    def __init__(self, name='glorot_normal', seed=None):
        super(GlorotNormal, self).__init__(name=name, seed=seed)


class RBFNormal(Initializer):
    """RBF网络的初始化器.
    """
    def __init__(self, name='rbf_normal', seed=None):
        super(RBFNormal, self).__init__(name=name, seed=seed)

    def __call__(self, hidden_units):
        """
        Arguments:
            hidden_units: int, 径向基函数网络的隐含层神经元数量.

        Notes:
            - 这里隐含层神经元中心本应用np.random.randn全部初始化,
              但是实际工程发现, 有负值的时候可能会导致求高斯函数的时候增加损失不收敛,
              因此, 全部初始化为正数.
        """
        parameters = {'w': np.zeros([1, hidden_units]),
                      'b': np.zeros([1, 1]),
                      'c': np.random.rand(hidden_units, 2),  # 隐含层神经元的中心
                      'beta': np.random.randn(1, hidden_units)}  # 高斯径向基函数的系数

        return parameters